"""
Common base class for line parser output data classes
"""
import re

from typing import List

from sys_toolkit.collection import CachedMutableSequence
from sys_toolkit.platform import detect_platform_family, detect_toolchain_family


class LineLoader(CachedMutableSequence):
    """
    Loader for line based data to cached mutable sequence
    """
    def __init__(self) -> None:
        super().__init__()
        self.__platform__ = detect_platform_family()
        self.__toolchain__ = detect_toolchain_family()
        self.__iter_items__ = None

    def __iter__(self):
        if not self.__loaded__:
            self.update()
        return iter(self.__items__)

    def __next__(self):
        if self.__iter_items__ is None:
            if not self.__loaded__:
                self.update()
            self.__iter_items__ = iter(self.__items__)
        try:
            return next(self.__iter_items__)
        except StopIteration as error:
            self.__iter_items__ = None
            raise StopIteration from error

    def __match_pattern_list__(self, lines: List[str], patterns: List[re.Pattern]) -> List[dict]:
        """
        Match lines to list of grouped regexp patterns, returning list
        of regexp groupdict matches.

        Lines must be iterable of strings and iterable of re.Pattern
        """
        assert isinstance(patterns, list)
        for pattern in patterns:
            assert isinstance(pattern, re.Pattern)
        matches = []
        for line in lines:
            assert isinstance(line, str)
            for pattern in patterns:
                match = pattern.match(line)
                if match:
                    matches.append(match.groupdict())
                    break
        return matches

    def update(self):
        """
        Require update to be implemented in child class
        """
        raise NotImplementedError




from pathlib import Path


# pylint: disable=too-few-public-methods
class MountpointOptions:
    """
    Options for filesystem mount point
    """
    def __init__(self, mountpoint, options):
        self.mountpoint = mountpoint
        options = self.__parse_options__(options)
        for flag in options:
            setattr(self, flag, True)

    @staticmethod
    def __parse_options__(options):
        """
        Parse options from string to a list
        """
        if isinstance(options, str):
            options = options.split(', ')
        return options


# pylint: disable=too-few-public-methods
class MountpointUsage:
    """

    Mountpoint usage stats data
    """
    def __init__(self, mountpoint):
        self.mountpoint = mountpoint
        self.size = None
        self.available = None
        self.used = None
        self.percent = None

    def __set_value__(self, attr, value):
        """
        Set value for usage counter
        """
        value = int(value)
        setattr(self, attr, value)

    def load_data(self, data):
        """
        Load usage data for mountpoint
        """
        for attr in ('size', 'available', 'used', 'percent'):
            assert attr in data
            self.__set_value__(attr, data[attr])


# pylint: disable=too-few-public-methods
class Filesystem:
    """
    Filesystem for a mountpoint
    """
    virtual_filesystems = ()

    def __init__(self, mountpoint, name):
        self.mountpoint = mountpoint
        self.name = name

    def __repr__(self):
        return self.name

    @property
    def is_virtual(self):
        """
        Return True if filesystem is a virtual filesystem
        """
        return self.name in self.virtual_filesystems


class Mountpoint:
    """
    Filesystem mount point linked to Mountpoints
    """
    filesystem_class = Filesystem
    options_class = MountpointOptions
    usage_class = MountpointUsage

    def __init__(self,
                 mountpoints,
                 device,
                 mountpoint,
                 filesystem=None,
                 options=None):

        self.mountpoints = mountpoints
        self.device = device
        self.mountpoint = mountpoint
        self.filesystem = self.filesystem_class(self, filesystem)
        self.options = self.options_class(self, options)
        self.usage = self.usage_class(self)

    def __repr__(self):
        return f'{self.device} mounted on {self.mountpoint}'

    @property
    def name(self):
        """
        Return basename of mountpoint
        """
        return Path(self.mountpoint).name

    @property
    def is_virtual(self):
        """
        Check if filesystem is virtual
        """
        return self.filesystem.is_virtual

    def load_usage_data(self, data):
        """
        Load filesystem usage data for mountpoint
        """
        self.usage.load_data(data)



"""
Platform specific code to handle fstab devices on OpenBSD
"""
from pathlib import Path
from typing import Optional

from sys_toolkit.collection import CachedMutableMapping
from sys_toolkit.subprocess import run_command


class DuidMap(CachedMutableMapping):
    """
    OpenBSD Disklabel Unique Identifiers (DUIDs) mapping to device names
    """
    def __get_sysctl_output__(self) -> str:
        """
        Get the sysctl output lines
        """
        stdout, _stderr = run_command(*('sysctl', '-n', 'hw.disknames'))
        return [str(line, 'utf-8') for line in stdout.splitlines()]

    def update(self, other=(), /, **kwds):
        """
        Update DUID map values
        """
        self.__items__ = {}
        self.__start_update__()
        for item in self.__get_sysctl_output__()[0].split(','):
            device, duid = item.split(':', 1)
            if not duid:
                duid = None
            self[device] = duid
        self.__finish_update__()

    def get_device(self, value: str) -> Optional[Path]:
        """
        Get device with full path by DUID. Return None if DUID is not found
        """
        for device, duid in self.items():
            if duid == value:
                return Path(f'/dev/{device}')
        return None



"""
Constants for fstab loading
"""

from pathlib import Path

FSTAB_PATH = Path('/etc/fstab')

FSTAB_FILE_NONE_VALUES = (
    'none',
)

FSTAB_FIELDS = (
    'fs_spec',
    'fs_file',
    'fs_vfstype',
    'fs_mntops',
    'fs_freq',
    'fs_passno',
)
FSTAB_INT_FIELDS = (
    'fs_freq',
    'fs_passno',
)



"""
MacOS Darwin mountpoints
"""

from .bsd import BSDMountpoint, BSDFilesystem, BSDMountPointOptions, BSDMountpointUsage

DARWIN_VIRTUAL_FILESYSTEMS = (
    'sysfs',
)


# pylint: disable=too-few-public-methods
class DarwinMountpointUsage(BSDMountpointUsage):
    """
    MacOS darwin specific mountpoint usage data
    """


# pylint: disable=too-few-public-methods
class DarwinMountPointOptions(BSDMountPointOptions):
    """
    MacOS darwin specific mountpoint options
    """


# pylint: disable=too-few-public-methods
class DarwinFilesystem(BSDFilesystem):
    """
    MacOS darwin specific mountpoint options
    """
    virtual_filesystems = DARWIN_VIRTUAL_FILESYSTEMS


class DarwinMountPoint(BSDMountpoint):
    """
    MacOS darwin specific mountpoint
    """
    filesystem_class = DarwinFilesystem
    options_class = DarwinMountPointOptions
    usage_class = DarwinMountpointUsage



"""
Generic BSD mountpoints (FreeBSD, MacOS)
"""

from .base import Mountpoint, Filesystem, MountpointOptions, MountpointUsage

BSD_VIRTUAL_FILESYSTEMS = (
    'devfs',
    'procfs',
)


# pylint: disable=too-few-public-methods
class BSDMountpointUsage(MountpointUsage):
    """
    BSD specific mountpoint usage data
    """
    def __init__(self, mountpoint):
        super().__init__(mountpoint)
        self.inodes_used = None
        self.inodes_available = None
        self.inodes_percent = None

    def load_data(self, data):
        """
        Load BSD specific filesystem usage data
        """
        super().load_data(data)
        for attr in ('inodes_used', 'inodes_available', 'inodes_percent'):
            if attr in data:
                self.__set_value__(attr, data[attr])


# pylint: disable=too-few-public-methods
class BSDMountPointOptions(MountpointOptions):
    """
    BSD specific mountpoint options
    """
    def __init__(self, mountpoint, options=None):
        options = self.__parse_options__(options)
        filesystem = options[0]
        options = options[1:]

        super().__init__(mountpoint, options)
        self.mountpoint.filesystem.name = filesystem


# pylint: disable=too-few-public-methods
class BSDFilesystem(Filesystem):
    """
    BSD specific mountpoint filesystem

    On BSD systems filesystem name comes from options
    """
    virtual_filesystems = BSD_VIRTUAL_FILESYSTEMS


class BSDMountpoint(Mountpoint):
    """
    BSD specific mountpoint
    """
    filesystem_class = BSDFilesystem
    options_class = BSDMountPointOptions
    usage_class = BSDMountpointUsage



"""
Mountpoints loader main class MountPoints()
"""
from typing import List

from sys_toolkit.subprocess import run_command

from ..base import LineLoader
from ..exceptions import FilesystemError
from .platform.base import Mountpoint
from .platform.bsd import BSDMountpoint
from .platform.darwin import DarwinMountPoint
from .platform.linux import LinuxMountPoint
from .platform.openbsd import OpenBSDMountPoint


from .constants import (
    GNU_MOUNT_COMMAND,
    GNU_DF_COMMAND,
    BSD_MOUNT_COMMAND,
    BSD_DF_COMMAND,
    RE_GNU_MOUNT_LINE,
    RE_GNU_DF_LINE,
    RE_BSD_MOUNT_LINE,
    RE_BSD_DF_LINE,
)


class Mountpoints(LineLoader):
    """
    Filesystem mount points with usage
    """
    __mount_command__ = None
    __df_command__ = None
    __re_mount_patterns__ = None
    __re_df_patterns__ = None

    def __init__(self):
        super().__init__()
        self.__detect_mountpoint_class__()
        self.__initialize_toolchain_based_data__()
        self.__iter_items__ = None

    def __detect_mountpoint_class__(self):
        """
        Detect mountpoint loader class based on platform family
        """
        if self.__platform__ == 'bsd':
            self.__mountpoint_class__ = BSDMountpoint
        elif self.__platform__ == 'darwin':
            self.__mountpoint_class__ = DarwinMountPoint
        elif self.__platform__ == 'linux':
            self.__mountpoint_class__ = LinuxMountPoint
        elif self.__platform__ == 'openbsd':
            self.__mountpoint_class__ = OpenBSDMountPoint
        else:
            raise FilesystemError(f'Unsupported OS platform: {self.__platform__}')

    def __initialize_toolchain_based_data__(self):
        """
        Initialize toolchain specific commands and regexp patterns
        """
        if self.__toolchain__ == 'bsd':
            self.__mount_command__ = BSD_MOUNT_COMMAND
            self.__df_command__ = BSD_DF_COMMAND
            self.__re_mount_patterns__ = RE_BSD_MOUNT_LINE
            self.__re_df_patterns__ = RE_BSD_DF_LINE

        elif self.__toolchain__ in ('gnu', 'openbsd'):
            self.__mount_command__ = GNU_MOUNT_COMMAND
            self.__df_command__ = GNU_DF_COMMAND
            self.__re_mount_patterns__ = RE_GNU_MOUNT_LINE
            self.__re_df_patterns__ = RE_GNU_DF_LINE

        else:
            raise FilesystemError(f'Unexpected toolchain detected: {self.__toolchain__}')

    def __get_mount_lines__(self) -> List[str]:
        """
        Return lines from mount command
        """
        stdout, _stderr = run_command(*self.__mount_command__)
        return [str(line, 'utf-8') for line in stdout.splitlines()]

    def __get_df_lines__(self) -> List[str]:
        """
        Return lines from df command
        """
        stdout, _stderr = run_command(*self.__df_command__)
        return [str(line, 'utf-8') for line in stdout.splitlines()]

    def __get_mountpoint_data__(self, lines: List[str]) -> List[dict]:
        """
        Return lines from mount command
        """
        return self.__match_pattern_list__(lines, self.__re_mount_patterns__)

    def __get_df_data__(self, lines: List[str]) -> List[dict]:
        """
        Return lines from df command
        """
        return self.__match_pattern_list__(lines, self.__re_df_patterns__)

    def append(self, value):
        assert isinstance(value, Mountpoint)
        return super().append(value)

    def insert(self, index, value):
        assert isinstance(value, Mountpoint)
        return super().insert(index, value)

    def update(self) -> None:
        """
        Get data for mountpoints
        """
        self.clear()
        self.__start_update__()
        mountpoints = {}
        for match in self.__get_mountpoint_data__(self.__get_mount_lines__()):
            item = self.__mountpoint_class__(self, **match)
            mountpoints[item.mountpoint] = item
            self.append(item)

        for match in self.__get_df_data__(self.__get_df_lines__()):
            item = mountpoints.get(match['mountpoint'], None)
            if item is not None:
                item.load_usage_data(match)
        self.__finish_update__()



"""
OpenBSD mounpoints
"""

from .bsd import BSDMountpoint, BSDFilesystem, BSDMountPointOptions, BSDMountpointUsage

OPENBSD_VIRTUAL_FILESYSTEMS = (
    'sysfs',
    'procfs',
)


# pylint: disable=too-few-public-methods
class OpenBSDMountpointUsage(BSDMountpointUsage):
    """
    OpenBSD specific mountpoint usage data
    """


# pylint: disable=too-few-public-methods
class OpenBSDMountPointOptions(BSDMountPointOptions):
    """
    OpenBSD specific mountpoint options
    """


# pylint: disable=too-few-public-methods
class OpenBSDFilesystem(BSDFilesystem):
    """
    OpenBSD specific mountpoint options
    """
    virtual_filesystems = OPENBSD_VIRTUAL_FILESYSTEMS


class OpenBSDMountPoint(BSDMountpoint):
    """
    OpenBSD specific mountpoint
    """
    filesystem_class = OpenBSDFilesystem
    options_class = OpenBSDMountPointOptions
    usage_class = OpenBSDMountpointUsage



"""
Linux mountpoints
"""
from .base import Mountpoint, Filesystem, MountpointOptions, MountpointUsage

LINUX_VIRTUAL_FILESYSTEMS = (
    'proc',
    'sysfs',
    'cgroup',
    'autofs',
    'hugetlbfs',
    'mqueue',
    'devpts',
    'devtmpfs',
    'tmpfs',
    'fusectl',
    'pstore',
    'configfs',
    'selinuxfs',
    'securityfs',
    'debugfs',
    'rpc_pipefs',
    'binfmt_misc',
    'nfsd',
    'fuse.vmware-vmblock',
)


# pylint: disable=too-few-public-methods
class LinuxMountpointUsage(MountpointUsage):
    """
    Linux specific mountpoint usage data
    """


# pylint: disable=too-few-public-methods
class LinuxMountPointOptions(MountpointOptions):
    """
    Linux specific mountpoint options
    """


# pylint: disable=too-few-public-methods
class LinuxFilesystem(Filesystem):
    """
    Linux specific mountpoint options
    """
    virtual_filesystems = LINUX_VIRTUAL_FILESYSTEMS


class LinuxMountPoint(Mountpoint):
    """
    Linux specific mountpoint
    """
    filesystem_class = LinuxFilesystem
    options_class = LinuxMountPointOptions
    usage_class = LinuxMountpointUsage



"""
Loader for fstab file details with OS specific
"""
from pathlib import Path
from operator import eq, ne, ge, gt, le, lt
from typing import List, Optional

from ..base import LineLoader
from ..exceptions import FilesystemError
from .constants import FSTAB_PATH, FSTAB_FIELDS, FSTAB_INT_FIELDS, FSTAB_FILE_NONE_VALUES


class FstabItem:
    """
    Generic fstab item text line
    """
    __line__: str

    def __init__(self, line: str) -> None:
        self.__line__ = line

    def __repr__(self) -> str:
        return str(self.__line__)

    def __compare__(self, other, operator):
        """
        Compare fstab items to string or FstabItem
        """
        if isinstance(other, str):
            return operator(self.__line__, other)
        if isinstance(other, FstabItem):
            return operator(self.__line__, other.__line__)
        raise TypeError(f'Error comparing FstabItem to {type(other)}: unexpected object')

    def __eq__(self, other):
        return self.__compare__(other, eq)

    def __ne__(self, other):
        return self.__compare__(other, ne)

    def __ge__(self, other):
        return self.__compare__(other, ge)

    def __gt__(self, other):
        return self.__compare__(other, gt)

    def __le__(self, other):
        return self.__compare__(other, le)

    def __lt__(self, other):
        return self.__compare__(other, lt)


# pylint: disable=too-few-public-methods
class FstabComment(FstabItem):
    """
    Comment line in fstab
    """


class FstabEntry(FstabItem):
    """
    Line in fstab
    """
    fs_spec: str
    fs_file: str
    fs_vfstype: str
    fs_mntops: str
    fs_freq: Optional[int]
    fs_freq: Optional[int]

    def __init__(self, line: str) -> None:
        super().__init__(line)
        self.__parse_line__(line)

    def __parse_line__(self, line: str) -> None:
        """
        Parse fstab line to entry attributes with FSTAB_FIELDS
        """
        fields = line.split()
        if len(fields) < 4 or len(fields) > 6:
            raise FilesystemError(f'Invalid fstab line: {line}')
        for index, field in enumerate(FSTAB_FIELDS):
            try:
                value = fields[index]
            except IndexError:
                value = None
            if field in FSTAB_INT_FIELDS and value is not None:
                value = int(value)
            setattr(self, field, value)

    @property
    def device(self) -> Optional[Path]:
        """
        Return path of device if fs_spec contains a path
        """
        return Path(self.fs_spec) if self.fs_spec[0] == '/' else None

    @property
    def label(self) -> Optional[str]:
        """
        Return label of device if fs_spec is in LABEL format
        """
        return self.fs_spec[5:] if self.fs_spec[:5].upper() == 'LABEL=' else None

    @property
    def mountpoint(self) -> Optional[Path]:
        """
        Return path of device if fs_spec contains a path
        """
        return Path(self.fs_file) if self.fs_file not in FSTAB_FILE_NONE_VALUES else None

    @property
    def options(self) -> List[str]:
        """
        Return filesystem options as list of strings
        """
        return self.fs_mntops.split(',')

    @property
    def partition_label(self) -> Optional[str]:
        """
        Return partition LABEL of device if fs_spec is in PARTLABEL format
        """
        return self.fs_spec[10:] if self.fs_spec[:10].upper() == 'PARTLABEL=' else None

    @property
    def partition_uuid(self) -> Optional[str]:
        """
        Return partition UUID of device if fs_spec is in PARTUUID format
        """
        return self.fs_spec[9:] if self.fs_spec[:9].upper() == 'PARTUUID=' else None

    @property
    def uuid(self) -> Optional[str]:
        """
        Return UUID of device if fs_spec is in UUID format
        """
        return self.fs_spec[5:] if self.fs_spec[:5].upper() == 'UUID=' else None

    @property
    def vfstypes(self) -> List[str]:
        """
        Return filesystem types as list of strings
        """
        return self.fs_vfstype.split(',')


class Fstab(LineLoader):
    """
    Lines in /etc/fstab
    """
    __lines__: list[str]

    def __init__(self, path: Optional[str] = None) -> None:
        super().__init__()
        self.path = Path(path) if path is not None else FSTAB_PATH
        self.__lines__ = []

    def __get_fstab_lines__(self) -> List[str]:
        """
        Load lines from fstab file as unicode strings
        """
        with self.path.open('r', encoding='utf-8') as handle:
            return handle.readlines()

    def __load_fstab__(self) -> None:
        """
        Load fstab lines. This is called from update() method only
        """
        self.__lines__ = []
        for line in self.__get_fstab_lines__():
            line = line.rstrip()
            if line == '' or line.startswith('#'):
                self.__lines__.append(FstabComment(line))
                continue

            entry = FstabEntry(line)
            self.__lines__.append(entry)
            self.append(entry)

    def update(self) -> None:
        """
        Update fstab information
        """
        self.clear()
        self.__start_update__()
        try:
            self.__load_fstab__()
        except FilesystemError as error:
            self.__reset__()
            raise FilesystemError(error) from error
        self.__finish_update__()

    def __encode_mountpoint_path__(self, path: str) -> Path:
        """
        Encode mountpoint path string. Replaces space and tab in string with
        escape sequences ’\040' and '\011' respectively as is done on Linux

        For full BSD compatibility this should implement full strunvis(3) encoding support
        """
        return Path(str(path).replace(' ', '\\040').replace('\t', '\\011'))

    def __get_by_attr__(self, attr: str, value: str) -> Optional[FstabEntry]:
        """
        Get an entry by property or attribute value. Returns None if no match is found
        """
        for item in self:
            if getattr(item, attr, None) == value:
                return item
        return None

    def get_by_uuid(self, uuid: str) -> Optional[FstabEntry]:
        """
        Get a fstab item by UUID
        """
        return self.__get_by_attr__('uuid', uuid)

    def get_by_label(self, label: str) -> Optional[FstabEntry]:
        """
        Get a fstab item by label
        """
        return self.__get_by_attr__('label', label)

    def get_by_device(self, path: str) -> Optional[FstabEntry]:
        """
        Get a fstab item by label
        """
        return self.__get_by_attr__('device', Path(path))

    def get_by_mountpoint(self, path: str) -> Optional[FstabEntry]:
        """
        Get a fstab item by label
        """
        return self.__get_by_attr__('mountpoint', self.__encode_mountpoint_path__(path))



"""
Constants for parsing OS family specific mount and usage line outputs
"""
import re

GNU_MOUNT_COMMAND = ('mount',)
# Patterns to match lines from GNU based 'mount' command
RE_GNU_MOUNT_LINE = [
    re.compile(
        # pylint: disable=consider-using-f-string
        r'^{}$'.format(
            r'\s+'.join([
                r'(?P<device>[^\s]*) on',
                r'(?P<mountpoint>[^\s]*) type',
                r'(?P<filesystem>[^\s]*)',
                r'\((?P<options>[^\)]*)\)'
            ])
        )
    )
]
BSD_MOUNT_COMMAND = ('mount',)
# Patterns to match lines from BSD 'mount' output
RE_BSD_MOUNT_LINE = [
    re.compile(
        # pylint: disable=consider-using-f-string
        r'^{}$'.format(
            r'\s+'.join([
                r'(?P<device>[^\s]*) on',
                r'(?P<mountpoint>[^\s]*)',
                r'\((?P<options>[^\)]*)\)'
            ])
        )
    )
]

GNU_DF_COMMAND = ('df', '-Pk')
# Patterns to match lines from GNU based 'df -Pk' output
RE_GNU_DF_LINE = [
    re.compile(
        # pylint: disable=consider-using-f-string
        r'^{}$'.format(
            r'\s+'.join([
                r'(?P<device>[^\s]*)',
                r'(?P<size>\d+)',
                r'(?P<used>\d+)',
                r'(?P<available>[\d-]+)',
                r'(?P<percent>\d+)%',
                r'(?P<mountpoint>.*)',
            ])
        )
    )
]

BSD_DF_COMMAND = ('df', '-Pki')
# Patterns to match lines from BSD based 'df -Pki' output
RE_BSD_DF_LINE = [
    re.compile(
        # pylint: disable=consider-using-f-string
        r'^{}$'.format(
            r'\s+'.join([
                r'(?P<device>[^\s]*)',
                r'(?P<size>\d+)',
                r'(?P<used>\d+)',
                r'(?P<available>\d+)',
                r'(?P<percent>\d+)%',
                r'(?P<inodes_used>\d+)',
                r'(?P<inodes_available>\d+)',
                r'(?P<inodes_percent>\d+)%',
                r'(?P<mountpoint>.*)',
            ])
        )
    )
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fs_toolkit',
 'fs_toolkit.fstab',
 'fs_toolkit.fstab.platform',
 'fs_toolkit.mounts',
 'fs_toolkit.mounts.platform']

package_data = \
{'': ['*']}

install_requires = \
['sys-toolkit>=2,<3']

setup_kwargs = {
    'name': 'fs-toolkit',
    'version': '1.0.3',
    'description': 'Classes for filesystem usage utilities',
    'long_description': "![Unit Tests](https://github.com/hile/fs-toolkit/actions/workflows/unittest.yml/badge.svg)\n![Style Checks](https://github.com/hile/fs-toolkit/actions/workflows/lint.yml/badge.svg)\n\n# Filesystem information tools\n\nThis module contains utilities to query local filesystem information as python objects,\nincluding mount points, disk usage and fstab contents.\n\nThis tool does similar things as `psutil` package: it may be better suited to\nyour use and has many features missing from this module.\n\n## Installing\n\n```bash\npip install fs-toolkit\n```\n\n## Basic examples\n\nSome usage examples\n\nFstab:\n\n```bash\nfrom fs_toolkit.fstab import Fstab\nfstab = Fstab()\nfstab.get_by_mountpoint('/var/my-secrets').uuid\n```\n\nMounts and df (linked together):\n\n```bash\nfrom fs_toolkit.mounts import Mountpoints\nprint('\\n'.join(f'{mp.usage.used:10} {mp.mountpoint}' for mp in Mountpoints()))\n```\n",
    'author': 'Ilkka Tuohela',
    'author_email': 'hile@iki.fi',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
